from datetime import datetime, timedelta
from typing import TYPE_CHECKING, Union

from clipped.compact.pydantic import (
    PYDANTIC_VERSION,
    StrictFloat,
    StrictInt,
    StrictStr,
    strict_str_validator,
)
from clipped.config.constants import PARAM_REGEX

if TYPE_CHECKING:
    from clipped.compact.pydantic import CallableGenerator


class RefField(StrictStr):
    if PYDANTIC_VERSION.startswith("2."):
        from pydantic_core import core_schema

        @classmethod
        def __get_pydantic_core_schema__(
            cls, *args, **kwargs
        ) -> core_schema.CoreSchema:
            from pydantic_core import core_schema

            def _validate(value):
                value = strict_str_validator(value)
                return cls.validate(value)

            return core_schema.no_info_after_validator_function(
                _validate,
                core_schema.str_schema(),
            )

    else:

        @classmethod
        def __get_validators__(cls) -> "CallableGenerator":
            yield strict_str_validator
            yield cls.validate

    @classmethod
    def validate(cls, value, **kwargs):
        if not isinstance(value, str):
            return value

        param = PARAM_REGEX.search(value)
        if not param:  # TODO: Fix error message
            raise ValueError(f"Value must be a reference, received `{value}` instead.")
        return value


BoolOrRef = Union[bool, RefField]
IntOrRef = Union[StrictInt, RefField]
StrictFloatOrRef = Union[StrictFloat, RefField]
FloatOrRef = Union[float, RefField]
DatetimeOrRef = Union[datetime, RefField]
TimeDeltaOrRef = Union[timedelta, RefField]



import logging
import os
import shutil
import tarfile
import tempfile

from contextlib import contextmanager
from datetime import datetime, timedelta
from typing import Any, List, Optional, Pattern, Tuple, Union

from clipped.utils.lists import to_list

_logger = logging.getLogger("clipped.utils.paths")


def check_or_create_path(path: Optional[str] = None, is_dir: bool = False):
    if not is_dir:
        path = os.path.dirname(os.path.abspath(path))
    if not os.path.exists(path):
        os.makedirs(path, exist_ok=True)


def delete_path(path: str, reraise: bool = False):
    if not os.path.exists(path):
        return
    try:
        if os.path.isfile(path):
            os.remove(path)
        else:
            shutil.rmtree(path)
    except OSError as e:
        if reraise:
            raise e
        _logger.warning("Could not delete path `%s`", path)


def create_path(path: str, reraise: bool = False) -> None:
    try:
        os.mkdir(path)
    except FileExistsError:
        pass
    except OSError as e:
        if reraise:
            raise e
        _logger.warning("Could not create path `%s`, exception %s", path, e)


def get_tmp_path(path: str) -> str:
    return os.path.join("/tmp", path)


def create_tmp_dir(dir_name: str) -> None:
    create_path(get_tmp_path(dir_name))


def delete_tmp_dir(dir_name: str) -> None:
    delete_path(get_tmp_path(dir_name))


def copy_to_tmp_dir(path: str, dir_name: str, reraise: bool = False) -> str:
    tmp_path = get_tmp_path(dir_name)
    if os.path.exists(tmp_path):
        return tmp_path
    try:
        shutil.copytree(path, tmp_path)
    except FileExistsError as e:
        if reraise:
            raise e
        _logger.warning("Path already exists `%s`, exception %s", path, e)
    return tmp_path


def copy_file(filename: str, path_to: str, use_basename: bool = True) -> Optional[str]:
    if use_basename:
        path_to = append_basename(path_to, filename)

    if filename == path_to:
        return

    check_or_create_path(path_to, is_dir=False)
    shutil.copy(filename, path_to)
    return path_to


@contextmanager
def get_files_by_paths(file_type: str, filepaths: List[str]) -> Tuple[List[str], int]:
    local_files = []
    total_file_size = 0

    for filepath in filepaths:
        local_files.append(
            (file_type, (unix_style_path(filepath), open(filepath, "rb"), "text/plain"))
        )
        total_file_size += os.path.getsize(filepath)

    yield local_files, total_file_size

    # close all files to avoid WindowsError: [Error 32]
    for f in local_files:
        f[1][1].close()


def get_files_and_dirs_in_path(
    path: str,
    exclude: Optional[List[str]] = None,
    collect_dirs: bool = False,
) -> Tuple[List[str], List[str]]:
    result_files = []
    result_dirs = []
    exclude = to_list(exclude, check_none=True)
    for root, dirs, files in os.walk(path, topdown=True):
        if exclude:
            dirs[:] = [d for d in dirs if d not in exclude]
        _logger.debug("Root:%s, Dirs:%s", root, dirs)
        for file_name in files:
            result_files.append(os.path.join(root, file_name))
        if collect_dirs:
            for dir_name in dirs:
                result_dirs.append(os.path.join(root, dir_name))
    return result_files, result_dirs


def get_files_in_path(path: str, exclude: Optional[List[str]] = None) -> List[str]:
    return get_files_and_dirs_in_path(path, exclude, False)[0]


def get_dirs_under_path(path: str) -> List[str]:
    return [
        name for name in os.listdir(path) if os.path.isdir(os.path.join(path, name))
    ]


def delete_old_files(path: str, hours: Optional[float] = 24) -> Tuple[int, List[str]]:
    """
    Delete files older than specified hours in the given directory and its subdirectories.

    Args:
        path (str): Directory path to start searching from
        hours (float): Number of hours, files older than this will be deleted

    Returns:
        tuple[int, list[str]]: Count of deleted files and list of deleted file paths
    """
    deleted_count = 0
    deleted_files = []
    current_time = datetime.now()
    cutoff_time = current_time - timedelta(hours=hours)

    files = get_files_in_path(path)
    for file_path in files:
        try:
            # Get the last modified time of the file
            file_time = datetime.fromtimestamp(os.path.getmtime(file_path))

            # Check if file is older than cutoff time
            if file_time < cutoff_time:
                os.remove(file_path)
                deleted_count += 1
                deleted_files.append(file_path)
        except (OSError, PermissionError) as e:
            _logger.debug(f"Error deleting old file {file_path}: {e}")

    return deleted_count, deleted_files


@contextmanager
def get_files_in_path_context(
    path: str, exclude: Optional[List[str]] = None
) -> List[str]:
    """
    Gets all the files under a certain path.

    Args:
        path: `str`. The path to traverse for collecting files.
        exclude: `list`. List of paths to excludes.

    Returns:
         list of files collected under the path.
    """
    yield get_files_in_path(path, exclude=exclude)


def unix_style_path(path) -> str:
    if os.path.sep != "/":
        return path.replace(os.path.sep, "/")
    return path


def create_tarfile(
    files: List[str], tar_path: str, relative_to: Optional[str] = None
) -> None:
    """Create a tar file based on the list of files passed"""
    with tarfile.open(tar_path, "w:gz") as tar:
        for f in files:
            arcname = os.path.relpath(f, relative_to) if relative_to else None
            tar.add(f, arcname=arcname)


@contextmanager
def create_tarfile_from_path(
    files: List[str], path_name: str, relative_to: Optional[str] = None
) -> str:
    """Create a tar file based on the list of files passed"""
    fd, filename = tempfile.mkstemp(prefix=path_name, suffix=".tar.gz")
    create_tarfile(files, filename, relative_to)
    yield filename

    # clear
    os.close(fd)
    os.remove(filename)


def untar_file(
    filename: Optional[str] = None,
    delete_tar: bool = True,
    extract_path: Optional[str] = None,
    use_filepath: bool = True,
):
    extract_path = extract_path or "."
    if use_filepath:
        extract_path = os.path.join(extract_path, filename.split(".tar.gz")[0])
    check_or_create_path(extract_path, is_dir=True)
    _logger.info("Untarring the contents of the file ...")
    # Untar the file
    with tarfile.open(filename) as tar:
        tar.extractall(extract_path)
    if delete_tar:
        _logger.info("Cleaning up the tar file ...")
        os.remove(filename)
    return extract_path


def move_recursively(src: str, dst: str):
    files = os.listdir(src)

    for f in files:
        shutil.move(os.path.join(src, f), dst)


def append_basename(path: str, filename: str):
    """
    Adds the basename of the filename to the path.

    Args:
        path: `str`. The path to append the basename to.
        filename: `str`. The filename to extract the base name from.

    Returns:
         str
    """
    return os.path.join(path, os.path.basename(filename))


def check_dirname_exists(path: str, is_dir: bool = False, reraise: bool = True):
    if not is_dir:
        path = os.path.dirname(os.path.abspath(path))
    if not os.path.isdir(path):
        error = "The parent path is not a directory {}".format(path)
        if reraise:
            raise OSError(error)
        _logger.warning(error)
        return False
    return True


def create_project_tmp(project_name: str) -> str:
    base_path = os.path.join("/tmp", project_name)
    if not os.path.exists(base_path):
        try:
            os.makedirs(base_path)
        except OSError:
            # Except permission denied and potential race conditions
            # in multi-threaded environments.
            _logger.warning("Could not create config directory `%s`", base_path)
    return base_path


def get_path_extension(filepath: str) -> str:
    return ".".join(os.path.basename(filepath).split(".")[1:]).lower()


def get_base_filename(filepath: str) -> str:
    return os.path.basename(filepath).split(".")[0]


def module_type(obj: Any, type_pattern: Union[str, Pattern]) -> bool:
    obj_type = type(obj)
    module = obj_type.__module__
    name = obj_type.__name__
    actual_fqn = "%s.%s" % (module, name)
    if isinstance(type_pattern, str):
        return type_pattern == actual_fqn
    else:
        return type_pattern.match(actual_fqn) is not None


def copy_file_path(from_path: str, asset_path: str):
    check_or_create_path(asset_path, is_dir=False)
    shutil.copy(from_path, asset_path)


def copy_dir_path(from_path: str, asset_path: str):
    check_or_create_path(asset_path, is_dir=False)
    shutil.copytree(from_path, asset_path)


def copy_file_or_dir_path(
    from_path: str, asset_path: str, use_basename: bool = False
) -> str:
    if use_basename:
        dir_name = os.path.basename(os.path.normpath(from_path))
        asset_path = (
            os.path.join(asset_path, dir_name) if asset_path is not None else dir_name
        )
    check_or_create_path(asset_path, is_dir=False)
    if os.path.isfile(from_path):
        try:
            shutil.copy(from_path, asset_path)
        except shutil.SameFileError:
            pass
    else:
        shutil.copytree(from_path, asset_path)

    return asset_path


def get_relative_path_to(base_path, paths: List[str]) -> List[str]:
    results = []
    if not paths:
        return results
    for d in paths:
        if not os.path.isabs(d):
            results.append(os.path.join(base_path, d))
        else:
            results.append(d)

    return results


RW_R_R_PERMISSIONS = 0o644


def set_permissions(
    path: str, permissions: int = RW_R_R_PERMISSIONS, reraise: bool = False
):
    try:
        os.chmod(path, permissions or RW_R_R_PERMISSIONS)
    except OSError as e:
        if reraise:
            raise e
        _logger.info("Could not set permissions `%s`", path)



import functools
import logging

from typing import Optional, Tuple

from clipped.utils.versions import compare_versions


class DeprecatedWarning(DeprecationWarning):
    def __init__(self, message, details: Optional[str] = None):
        self.message = message
        self.details = details
        super(DeprecatedWarning, self).__init__(message, details)

    def __str__(self):
        return "%s\n%s" % (self.message, self.details)


class UnsupportedWarning(DeprecatedWarning):
    pass


def check_deprecation(
    current_version: Optional[str] = None,
    deprecation_version: Optional[str] = None,
    latest_version: Optional[str] = None,
) -> Tuple[bool, bool]:
    if deprecation_version is None and latest_version is not None:
        raise TypeError(
            "Cannot set `latest_version` to a value without also setting `deprecation_version`"
        )

    is_deprecated = False
    is_unsupported = False

    if current_version:
        if latest_version and compare_versions(current_version, latest_version, ">="):
            is_unsupported = True
        elif deprecation_version and compare_versions(
            current_version, deprecation_version, ">="
        ):
            is_deprecated = True
    else:
        # Automatically deprecate based if only deprecation version is provided.
        is_deprecated = True

    return is_deprecated, is_unsupported


def get_deprecation_warning_message(
    deprecation_version: str,
    latest_version: str,
    current_logic: str,
    new_logic: Optional[str] = None,
) -> str:
    message = [f"`{current_logic}` is deprecated as of `{deprecation_version}`"]
    if latest_version:
        message.append(f"it will be removed in `{latest_version}`")
    if new_logic:
        message.append(f"please use `{new_logic}` instead")
    return ", ".join(message) + "."


def warn_deprecation(
    deprecation_version: Optional[str] = None,
    latest_version: Optional[str] = None,
    current_version: Optional[str] = None,
    current_logic: Optional[str] = None,
    new_logic: Optional[str] = None,
    details: Optional[str] = None,
):
    is_deprecated, is_unsupported = check_deprecation(
        current_version=current_version,
        deprecation_version=deprecation_version,
        latest_version=latest_version,
    )

    if is_deprecated or is_unsupported:
        if is_unsupported:
            cls = UnsupportedWarning
        else:
            cls = DeprecatedWarning

        message = get_deprecation_warning_message(
            deprecation_version=deprecation_version,
            latest_version=latest_version,
            current_logic=current_logic,
            new_logic=new_logic,
        )
        logging.warning(cls(message, details))


def deprecated(
    deprecation_version: Optional[str] = None,
    latest_version: Optional[str] = None,
    current_version: Optional[str] = None,
    current_logic: Optional[str] = None,
    new_logic: Optional[str] = None,
    details: Optional[str] = None,
):
    """This decorator can be used to warn about deprecated functions.

    Example:
        # Class function
        class MyClass:
            @deprecated(deprecation_version=..., ...)
            def foo(self, a, b):
                ...

        # Function with other decorators
        @other_decorators_must_be_upper
        @deprecated(deprecation_version=..., ...)
        def my_func():
            pass
    """

    def wrapper(func):
        @functools.wraps(func)
        def _inner(*args, **kwargs):
            _current_logic = (
                current_logic
                or "The function `%(name)s` in file `%(filename)s` "
                "line number `%(line)s` is deprecated."
                % {
                    "name": func.__name__,
                    "filename": func.__code__.co_filename,
                    "line": func.__code__.co_firstlineno + 1,
                }
            )
            warn_deprecation(
                current_version=current_version,
                deprecation_version=deprecation_version,
                latest_version=latest_version,
                current_logic=_current_logic,
                new_logic=new_logic,
                details=details,
            )
            return func(*args, **kwargs)

        return _inner

    return wrapper



from typing import Dict, Optional

try:
    import requests
except ImportError:
    raise ImportError("This module depends on requests.")


def create_session(
    session: Optional[requests.Session] = None,
    session_attrs: Optional[Dict] = None,
) -> requests.Session:
    session = session or requests.Session()
    if not session_attrs:
        return session
    if "proxies" in session_attrs:
        session.proxies = session_attrs.pop("proxies")
    elif "proxy" in session_attrs:
        session.proxies = session_attrs.pop("proxy")
    if "stream" in session_attrs:
        session.stream = session_attrs.pop("stream")
    if "verify" in session_attrs or "verify_ssl" in session_attrs:
        session.verify = session_attrs.pop(
            "verify", session_attrs.pop("verify_ssl", True)
        )
    if "cert" in session_attrs:
        session.cert = session_attrs.pop("cert")
    if "max_redirects" in session_attrs:
        session.max_redirects = session_attrs.pop("max_redirects")
    if "trust_env" in session_attrs:
        session.trust_env = session_attrs.pop("trust_env")

    return session


def safe_request(
    url: str,
    method: str = None,
    params: Optional[Dict] = None,
    data: Optional[Dict] = None,
    json: Optional[Dict] = None,
    headers: Optional[Dict] = None,
    allow_redirects: bool = False,
    timeout: int = 30,
    verify_ssl: bool = True,
    session: Optional[requests.Session] = None,
    session_attrs: Optional[Dict] = None,
) -> requests.Response:
    """A slightly safer version of `request`."""

    session = create_session(session, session_attrs)

    kwargs = {}

    if json:
        kwargs["json"] = json
        if not headers:
            headers = {}
        headers.setdefault("Content-Type", "application/json")

    if data:
        kwargs["data"] = data

    if params:
        kwargs["params"] = params

    if headers:
        kwargs["headers"] = headers

    method = method or ("POST" if (data or json) else "GET")

    response = session.request(
        method=method,
        url=url,
        allow_redirects=allow_redirects,
        timeout=timeout,
        verify=verify_ssl,
        **kwargs
    )

    return response



from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Type, TypeVar

from clipped.utils.json import orjson_dumps, orjson_loads

T = TypeVar("T")
from pydantic.version import VERSION as PYDANTIC_VERSION

if PYDANTIC_VERSION.startswith("2."):
    from pydantic import AnyUrl, BaseConfig
    from pydantic import BaseModel as _BaseModel
    from pydantic import ByteSize, ConfigDict, Field
    from pydantic import FieldValidationInfo as ModelField
    from pydantic import (
        FiniteFloat,
        FutureDate,
        Json,
        NegativeFloat,
        NegativeInt,
        NonNegativeFloat,
        NonNegativeInt,
        NonPositiveFloat,
        NonPositiveInt,
        PastDate,
        PaymentCardNumber,
        PositiveFloat,
        PositiveInt,
        PrivateAttr,
        RootModel,
        SecretBytes,
        SecretStr,
        StrictBool,
        StrictBytes,
        StrictFloat,
        StrictInt,
        StrictStr,
        ValidationError,
        ValidationInfo,
        constr,
        create_model,
        field_validator,
        model_validator,
        validate_call,
    )
    from pydantic.deprecated.parse import Protocol, load_str_bytes
    from pydantic.deprecated.tools import NameFactory
    from pydantic.v1.datetime_parse import parse_date, parse_datetime, parse_duration
    from pydantic.v1.validators import strict_str_validator, uuid_validator

    validation_before = {"mode": "before"}
    validation_after = {"mode": "after"}
    validation_always = {}

    NAME_REGEX = r"^[-a-zA-Z0-9_]+\z"
    FULLY_QUALIFIED_NAME_REGEX = r"^[-a-zA-Z0-9_]+(:[-a-zA-Z0-9_.]+)?\z"

    def patter_constr(pattern):
        return constr(pattern=pattern)

    class PydanticConfig:
        populate_by_name = True
        validate_assignment = True
        arbitrary_types_allowed = True
        use_enum_values = True
        extra = "forbid"
        ser_json_timedelta = "float"
        protected_namespaces = ()

    class PydanticAllowConfig(PydanticConfig):
        extra = "allow"

    class PydanticAlwaysConfig(PydanticConfig):
        revalidate_instances = True

    def model_rebuild(model, **localns: Any) -> None:
        return model.model_rebuild(force=True, _types_namespace=localns)

    class PydanticTypeError(TypeError):
        pass

    class PydanticValueError(ValueError):
        pass

    def parse_obj_as(type_: Type[T], obj: Any, **type_forwarding) -> T:
        from pydantic import TypeAdapter

        adapter = TypeAdapter(type_)
        return adapter.validate_python(obj)

    class BaseModel(_BaseModel):
        def __init__(self, **data):
            if (
                hasattr(self, "_USE_DISCRIMINATOR")
                and hasattr(self, "_IDENTIFIER")
                and self._USE_DISCRIMINATOR
            ):
                data["kind"] = data.pop("kind", self._IDENTIFIER)
            super().__init__(**data)

        def model_dump(self, *args, **kwargs) -> Dict:
            # Handle custom fields
            exclude_fields = getattr(self, "_CUSTOM_DUMP_FIELDS", None)
            data = super().model_dump(*args, exclude=exclude_fields, **kwargs)
            if hasattr(self, "_dump_obj"):
                data = self._dump_obj(data)
            return data

        def dict(self, *args, **kwargs) -> Dict:
            return self.model_dump(*args, **kwargs)

    if TYPE_CHECKING:
        from pydantic.typing import CallableGenerator
else:
    from pydantic import AnyUrl, BaseConfig
    from pydantic import BaseModel as _BaseModel
    from pydantic import (
        ByteSize,
        Field,
        FiniteFloat,
        FutureDate,
        Json,
        NegativeFloat,
        NegativeInt,
        NonNegativeFloat,
        NonNegativeInt,
        NonPositiveFloat,
        NonPositiveInt,
        PastDate,
        PaymentCardNumber,
        PositiveFloat,
        PositiveInt,
        PrivateAttr,
        PydanticTypeError,
        PydanticValueError,
        SecretBytes,
        SecretStr,
        StrictBool,
        StrictBytes,
        StrictFloat,
        StrictInt,
        StrictStr,
        ValidationError,
        constr,
        create_model,
    )
    from pydantic import root_validator as model_validator
    from pydantic import validate_arguments as validate_call
    from pydantic import validator as field_validator
    from pydantic.datetime_parse import parse_date, parse_datetime, parse_duration
    from pydantic.fields import ModelField
    from pydantic.parse import Protocol, load_str_bytes
    from pydantic.tools import NameFactory, _get_parsing_type
    from pydantic.validators import strict_str_validator, uuid_validator

    validation_before = {"pre": True}
    validation_after = {"pre": False}
    validation_always = {"always": True}

    NAME_REGEX = r"^[-a-zA-Z0-9_]+\Z"
    FULLY_QUALIFIED_NAME_REGEX = r"^[-a-zA-Z0-9_]+(:[-a-zA-Z0-9_.]+)?\Z"

    def patter_constr(pattern):
        return constr(regex=pattern)

    def parse_obj_as(type_: Type[T], obj: Any, **type_forwarding) -> T:
        model_type = _get_parsing_type(
            type_,
        )
        model_type.update_forward_refs(**type_forwarding)
        return model_type(__root__=obj).__root__

    class PydanticConfig:
        allow_population_by_field_name = True
        validate_assignment = True
        arbitrary_types_allowed = True
        use_enum_values = True
        extra = "forbid"
        json_dumps: Callable[..., str] = orjson_dumps
        json_loads = orjson_loads

    class PydanticAllowConfig(PydanticConfig):
        extra = "allow"

    def model_rebuild(model, **localns: Any) -> None:
        return model.update_forward_refs(**localns)

    class RootModel(_BaseModel):
        pass

    class BaseModel(_BaseModel):
        def __init__(self, **data):
            if (
                hasattr(self, "_USE_DISCRIMINATOR")
                and hasattr(self, "_IDENTIFIER")
                and self._USE_DISCRIMINATOR
            ):
                data["kind"] = data.pop("kind", self._IDENTIFIER)
            super().__init__(**data)

        @property
        def model_fields(self):
            return self.__fields__

        @classmethod
        def model_construct(cls, *args, **data):
            return cls.construct(*args, **data)

        def dict(self, *args, **kwargs) -> Dict:
            data = super().dict(*args, **kwargs)
            if hasattr(self, "_dump_obj"):
                data = self._dump_obj(data)
            return data

        def model_dump(self, *args, **kwargs) -> Dict:
            return self.dict(*args, **kwargs)

        def model_dump_json(self, *args, **kwargs) -> str:
            return self.json(*args, **kwargs)

        @classmethod
        def model_validate(cls, *args, **data) -> "Model":
            return cls.parse_obj(*args, **data)

        @classmethod
        def model_rebuild(cls, **localns: Any) -> None:
            return cls.update_forward_refs(**localns)

    if TYPE_CHECKING:
        from pydantic.typing import CallableGenerator



import os
import signal

from polyaxon.logger import logger


def clean_process(pid: int) -> bool:
    try:
        logger.debug("Killing process with pid %s" % pid)
        try:
            if os.name == "nt":
                os.kill(pid, signal.CTRL_BREAK_EVENT)
            else:
                os.kill(pid, signal.SIGKILL)
            return True
        except Exception:
            # Windows process killing is flaky
            pass
    except Exception as e:
        logger.debug(e)
        return False
